@@ -1,9 +1,9 @@
 /*
- * $Id: dialog.c,v 1.231 2013/09/02 17:20:09 tom Exp $
+ * $Id: dialog.c,v 1.237 2014/02/19 19:49:04 tom Exp $
  *
  *  cdialog - Display simple dialog boxes from shell scripts
  *
- *  Copyright 2000-2012,2013	Thomas E. Dickey
+ *  Copyright 2000-2013,2014	Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -375,6 +375,7 @@ unescape_argv(int *argcp, char ***argvp)
 		fprintf(stderr, " arg%d:%s\n", k, (*argvp)[k]);
 	    }
 	    changed = dlg_eat_argv(argcp, argvp, j, 1);
+	    --j;
 	} else if (!strcmp((*argvp)[j], "--file")) {
 	    if (++count_includes > limit_includes)
 		dlg_exiterr("Too many --file options");
@@ -431,9 +432,16 @@ unescape_argv(int *argcp, char ***argvp)
 			    dialog_opts = dlg_realloc(bool, need, dialog_opts);
 			    assert_ptr(dialog_opts, "unescape_argv");
 			}
-			for (n = *argcp; n >= j + 2; --n) {
-			    (*argvp)[n + added - 2] = (*argvp)[n];
-			    dialog_opts[n + added - 2] = dialog_opts[n];
+			if (added > 2) {
+			    for (n = *argcp; n >= j + 2; --n) {
+				(*argvp)[n + added - 2] = (*argvp)[n];
+				dialog_opts[n + added - 2] = dialog_opts[n];
+			    }
+			} else if (added == 1) {
+			    for (n = j; n < *argcp; ++n) {
+				(*argvp)[n] = (*argvp)[n + 1];
+				dialog_opts[n] = dialog_opts[n + 1];
+			    }
 			}
 			for (n = 0; n < added; ++n) {
 			    (*argvp)[n + j] = list[n];
@@ -441,6 +449,7 @@ unescape_argv(int *argcp, char ***argvp)
 			}
 			*argcp += added - 2;
 			free(list);
+			--j;	/* force rescan */
 		    }
 		} else {
 		    dlg_exiterr("Cannot open --file %s", filename);
@@ -494,7 +503,7 @@ isOption(const char *arg)
 		}
 	    }
 	} else if (!strncmp(arg, "--", (size_t) 2) && isalpha(UCH(arg[2]))) {
-	    if (strlen(arg) == strspn(arg, OptionChars)) {
+	    if (strlen(arg) == (strspn) (arg, OptionChars)) {
 		result = TRUE;
 	    } else {
 		dlg_exiterr("Invalid option \"%s\"", arg);
@@ -1276,7 +1285,7 @@ Help(void)
     static const char *const tbl_1[] =
     {
 	"cdialog (ComeOn Dialog!) version %s",
-	"Copyright 2000-2012,2013 Thomas E. Dickey",
+	"Copyright 2000-2013,2014 Thomas E. Dickey",
 	"This is free software; see the source for copying conditions.  There is NO",
 	"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.",
 	"",
@@ -1854,10 +1863,20 @@ main(int argc, char *argv[])
 	    Usage("Expected a box option");
 	}
 
-	if (lookupOption(argv[offset], 2) != o_checklist
-	    && dialog_vars.separate_output) {
-	    sprintf(temp, "Expected --checklist, not %.20s", argv[offset]);
-	    Usage(temp);
+	if (dialog_vars.separate_output) {
+	    switch (lookupOption(argv[offset], 2)) {
+#ifdef HAVE_XDIALOG2
+	    case o_buildlist:
+	    case o_treeview:
+#endif
+	    case o_checklist:
+		break;
+	    default:
+		sprintf(temp,
+			"Unexpected widget with --separate-output %.20s",
+			argv[offset]);
+		Usage(temp);
+	    }
 	}
 
 	if (dialog_state.aspect_ratio == 0)
